#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <mntent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

typedef struct sProperty
{
	char *name;
	char *value;
} Property;

typedef struct sDevice
{
	char *node;
	char *label;
	char *description;
	int mounted;
	time_t time;
} Device;

int verbosity = 0;

int parse_property(char *str, int size, Property *prop)
{
	int equals = -1;
	int i;

	for(i=0; (equals<0 && i<size); ++i)
		if(str[i]=='=')
			equals = i;

	if(equals<0)
		return -1;

	prop->name = malloc(equals+1);
	strncpy(prop->name, str, equals);
	prop->name[equals] = 0;

	prop->value = malloc(size-equals);
	strncpy(prop->value, str+equals+1, size-equals-1);
	prop->value[size-equals-1] = 0;

	return 0;
}

Property *get_device_properties(char *node)
{
	int pid;
	int pipe_fd[2];

	pipe(pipe_fd);

	pid = fork();
	if(pid==0)
	{
		if(verbosity>=2)
			printf("Running udevadm info -q property -n \"%s\"\n", node);

		close(pipe_fd[0]);
		dup2(pipe_fd[1], 1);

		execl("/sbin/udevadm", "udevadm", "info", "-q", "property", "-n", node, NULL);
		_exit(1);
	}
	else if(pid>0)
	{
		char *buf;
		int bufsize;
		int pos = 0;
		int eof = 0;
		Property *props = NULL;
		int n_props = 0;

		close(pipe_fd[1]);

		bufsize = 256;
		buf = (char *)malloc(bufsize);

		while(1)
		{
			int newline;
			int i;
			Property prop;

			if(!eof)
			{
				int len;

				len = read(pipe_fd[0], buf+pos, bufsize-pos);
				if(len==0)
					eof = 1;
				else if(len==-1)
					break;
				pos += len;
			}

			newline = -1;
			for(i=0; (newline<0 && i<pos); ++i)
				if(buf[i]=='\n')
					newline = i;

			if(newline<0)
			{
				if(eof)
					break;
				bufsize *= 2;
				buf = (char *)realloc(buf, bufsize);
				continue;
			}

			if(parse_property(buf, newline, &prop)==0)
			{
				props = (Property *)realloc(props, (n_props+2)*sizeof(Property));
				props[n_props] = prop;
				++n_props;

				memmove(buf, buf+newline+1, pos-newline-1);
				pos -= newline+1;
			}
			else
				break;
		}

		free(buf);

		if(props)
		{
			props[n_props].name = NULL;
			props[n_props].value = NULL;
		}

		waitpid(pid, NULL, 0);
		close(pipe_fd[0]);

		return props;
	}
	else
	{
		close(pipe_fd[0]);
		close(pipe_fd[1]);

		return NULL;
	}
}

char *get_property_value(Property *props, char *name)
{
	int i;
	for(i=0; props[i].name; ++i)
		if(strcmp(props[i].name, name)==0)
			return props[i].value;
	return NULL;
}

int match_property_value(Property *props, char *name, char *value)
{
	char *v = get_property_value(props, name);
	if(!v)
		return value==NULL;
	return strcmp(v, value)==0;
}

void free_properties(Property *props)
{
	int i;
	if(!props)
		return;
	for(i=0; props[i].name; ++i)
	{
		free(props[i].name);
		free(props[i].value);
	}
	free(props);
}

char **get_mounted_devices(void)
{
	FILE *mtab;
	struct mntent *me;
	char **mounted = NULL;
	int n_mounted = 0;

	mtab = setmntent("/etc/mtab", "r");
	if(!mtab)
		return NULL;

	while((me = getmntent(mtab)))
	{
		mounted = (char **)realloc(mounted, (n_mounted+2)*sizeof(char *));
		mounted[n_mounted] = strdup(me->mnt_fsname);
		++n_mounted;
	}

	endmntent(mtab);
	mounted[n_mounted] = NULL;

	return mounted;
}

int is_mounted(char **mounted, char *devname)
{
	int i;
	for(i=0; mounted[i]; ++i)
		if(!strcmp(devname, mounted[i]))
			return 1;
	return 0;
}

void free_mounted_devices(char **mounted)
{
	int i;
	if(!mounted)
		return;
	for(i=0; mounted[i]; ++i)
		free(mounted[i]);
	free(mounted);
}

int is_removable(char *devpath)
{
	char fnbuf[256];
	int len;
	char *ptr;
	int fd;

	len = snprintf(fnbuf, sizeof(fnbuf), "/sys%s", devpath);
	if(len+10>=(int)sizeof(fnbuf))
		return 0;

	for(ptr=fnbuf+len; (ptr>fnbuf && *ptr!='/'); --ptr) ;
	strcpy(ptr, "/removable");
	fd = open(fnbuf, O_RDONLY);
	if(fd!=-1)
	{
		char c;
		read(fd, &c, 1);
		close(fd);
		if(c=='1')
		{
			if(verbosity>=2)
				printf("  Removable\n");
			return 1;
		}
		if(verbosity>=2)
			printf("  Not removable\n");
	}

	return 0;
}

int check_buses(char *devpath, char **buses)
{
	char fnbuf[256];
	char *ptr;
	int len;

	len = snprintf(fnbuf, sizeof(fnbuf), "/sys%s", devpath);
	if(len+10>=(int)sizeof(fnbuf))
		return 0;

	for(ptr=fnbuf+len; ptr>fnbuf+12; --ptr)
		if(*ptr=='/')
		{
			char linkbuf[256];
			strcpy(ptr, "/subsystem");
			len = readlink(fnbuf, linkbuf, sizeof(linkbuf)-1);
			*ptr = 0;

			if(len!=-1)
			{
				int i;
				linkbuf[len] = 0;
				for(; (len>0 && linkbuf[len-1]!='/'); --len) ;
				if(verbosity>=2)
					printf("  Subsystem of %s is %s\n", fnbuf, linkbuf+len);
				for(i=0; buses[i]; ++i)
					if(strcmp(linkbuf+len, buses[i])==0)
						return 1;
			}
		}

	return 0;
}

int can_mount(Property *props)
{
	static char *removable_buses[] = { "usb", "firewire", 0 };
	char *devpath;
	int i;

	if(!match_property_value(props, "DEVTYPE", "partition"))
		return 0;

	devpath = get_property_value(props, "DEVPATH");
	if(is_removable(devpath))
		return 1;

	for(i=0; removable_buses[i]; ++i)
		if(match_property_value(props, "ID_BUS", removable_buses[i]))
			return 1;

	return check_buses(devpath, removable_buses);
}

char **get_device_nodes(char *dirname)
{
	DIR *dir;
	struct dirent *de;
	char fnbuf[256];
	char linkbuf[256];
	struct stat st;
	char **nodes = NULL;
	int n_nodes = 0;
	char **checked = NULL;
	int n_checked = 0;
	int i;

	dir = opendir(dirname);
	if(!dir)
		return NULL;

	while((de = readdir(dir)))
	{
		char *node;
		int duplicate = 0;

		if(de->d_name[0]=='.' && (de->d_name[1]==0 || (de->d_name[1]=='.' && de->d_name[2]==0)))
			continue;

		snprintf(fnbuf, sizeof(fnbuf), "%s/%s", dirname, de->d_name);

		node = fnbuf;
		lstat(fnbuf, &st);
		if(S_ISLNK(st.st_mode))
		{
			int len;
			len = readlink(fnbuf, linkbuf, sizeof(linkbuf)-1);
			if(len!=-1)
			{
				linkbuf[len] = 0;
				node = linkbuf;
			}
		}

		if(checked)
		{
			for(i=0; (!duplicate && i<n_checked); ++i)
				if(strcmp(node, checked[i])==0)
					duplicate = 1;
		}
		if(duplicate)
		{
			if(verbosity>=2)
				printf("Device %s is a duplicate\n", fnbuf);
			continue;
		}

		checked = (char **)realloc(checked, (n_checked+1)*sizeof(char *));
		checked[n_checked] = strdup(node);
		++n_checked;

		nodes = (char **)realloc(nodes, (n_nodes+2)*sizeof(char *));
		nodes[n_nodes] = strdup(fnbuf);
		++n_nodes;
	}

	closedir(dir);
	if(checked)
	{
		for(i=0; i<n_checked; ++i)
			free(checked[i]);
		free(checked);
	}

	if(nodes)
		nodes[n_nodes] = NULL;

	return nodes;
}

Device *get_devices(void)
{
	char **nodes = NULL;
	Device *devices = NULL;
	int n_devices = 0;
	char **mounted = NULL;
	int i;

	nodes = get_device_nodes("/dev/disk/by-id");
	mounted = get_mounted_devices();

	for(i=0; nodes[i]; ++i)
	{
		if(verbosity>=1)
			printf("Examining device %s\n", nodes[i]);

		Property *props = get_device_properties(nodes[i]);
		if(!props)
		{
			if(verbosity>=2)
				printf("  No properties\n");
			continue;
		}

		if(verbosity>=2)
		{
			int j;
			for(j=0; props[j].name; ++j)
				printf("  %s = %s\n", props[j].name, props[j].value);
		}

		if(can_mount(props))
		{
			char *devname;
			char *label;
			char *vendor;
			char *model;
			char buf[256];
			char pos;
			struct stat st;

			if(verbosity>=1)
				printf("  Using device\n");

			devname = get_property_value(props, "DEVNAME");

			label = get_property_value(props, "ID_FS_LABEL");
			if(!label)
				label = get_property_value(props, "ID_FS_UUID");
			if(!label)
			{
				char *ptr;

				label = devname;
				for(ptr=label; *ptr; ++ptr)
					if(*ptr=='/')
						label = ptr+1;
			}

			vendor = get_property_value(props, "ID_VENDOR");
			model = get_property_value(props, "ID_MODEL");

			pos = snprintf(buf, sizeof(buf), "%s", label);
			if(vendor && model)
				pos += snprintf(buf+pos, sizeof(buf)-pos, " (%s %s)", vendor, model);

			stat(nodes[i], &st);

			devices = (Device *)realloc(devices, (n_devices+2)*sizeof(Device));
			devices[n_devices].node = nodes[i];
			devices[n_devices].label = strdup(label);
			devices[n_devices].description = strdup(buf);
			devices[n_devices].mounted = is_mounted(mounted, devname);
			devices[n_devices].time = st.st_mtime;
			++n_devices;
		}
		else
			free(nodes[i]);
		free_properties(props);
	}

	free(nodes);
	free_mounted_devices(mounted);

	if(devices)
	{
		devices[n_devices].node = NULL;
		devices[n_devices].label = NULL;
		devices[n_devices].description = NULL;
	}

	return devices;
}

void free_devices(Device *devices)
{
	int i;
	if(!devices)
		return;
	for(i=0; devices[i].node; ++i)
	{
		free(devices[i].node);
		free(devices[i].label);
		free(devices[i].description);
	}
	free(devices);
}

void row_activated(GtkTreeView *list, GtkTreePath *path, GtkTreeViewColumn *column, gpointer user_data)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	int umount = *(int *)user_data;

	model = gtk_tree_view_get_model(list);

	if(gtk_tree_model_get_iter(model, &iter, path))
	{
		Device *device;
		int pid;
		int pipe_fd[2];

		gtk_tree_model_get(model, &iter, 1, &device, -1);

		pipe(pipe_fd);

		pid = fork();
		if(pid==0)
		{
			if(verbosity>=1)
			{
				if(umount)
					printf("Running pumount %s\n", device->node);
				else
					printf("Running pmount %s %s\n", device->node, device->label);
			}

			close(pipe_fd[0]);
			dup2(pipe_fd[1], 1);
			dup2(pipe_fd[1], 2);

			if(umount)
				execl("/usr/bin/pumount", "pumount", device->node, NULL);
			else
				execl("/usr/bin/pmount", "pmount", device->node, device->label, NULL);
			_exit(1);
		}
		else if(pid>0)
		{
			char buf[1024];
			int pos = 0;
			int status;

			close(pipe_fd[1]);

			while(1)
			{
				int len;

				len = read(pipe_fd[0], buf+pos, sizeof(buf)-pos-1);
				if(len<=0)
					break;
				pos += len;
			}

			buf[pos] = 0;

			waitpid(pid, &status, 0);
			if(!WIFEXITED(status) || WEXITSTATUS(status))
			{
				GtkWidget *dialog;

				dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "%s", buf);
				g_signal_connect(dialog, "response", &gtk_main_quit, NULL);
				gtk_widget_show_all(dialog);
			}
			else
				gtk_main_quit();
		}
		else
		{
		}
	}

	(void)column;
}

gboolean key_press(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	if(event->key.keyval==GDK_KEY_Escape)
	{
		gtk_main_quit();
		return TRUE;
	}

	(void)widget;
	(void)user_data;

	return FALSE;
}

int main(int argc, char **argv)
{
	GtkWidget *window;
	GtkWidget *viewport;
	GtkWidget *list;
	GtkListStore *store;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	Device *devices;
	int i;
	time_t latest;
	int opt;
	int umount = 0;
	int n_listed;

	gtk_init(&argc, &argv);

	while((opt = getopt(argc, argv, "vu"))!=-1) switch(opt)
	{
	case 'v':
		++verbosity;
		break;
	case 'u':
		umount = 1;
		break;
	}

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);
	g_signal_connect(window, "destroy", G_CALLBACK(&gtk_main_quit), NULL);
	g_signal_connect(window, "key-press-event", G_CALLBACK(&key_press), NULL);

	viewport = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type(GTK_VIEWPORT(viewport), GTK_SHADOW_IN);
	gtk_container_add(GTK_CONTAINER(window), viewport);

	list = gtk_tree_view_new();
	gtk_container_add(GTK_CONTAINER(viewport), list);
	g_signal_connect(list, "row-activated", G_CALLBACK(&row_activated), &umount);

	store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_POINTER);
	gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(list),
		-1, "Device", gtk_cell_renderer_text_new(), "text", 0, NULL);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(list));

	devices = get_devices();
	n_listed = 0;
	if(devices)
	{
		latest = 0;
		for(i=0; devices[i].node; ++i)
			if(!devices[i].mounted==!umount)
			{
				gtk_list_store_append(store, &iter);
				gtk_list_store_set(store, &iter, 0, devices[i].description, 1, &devices[i], -1);
				if(devices[i].time>latest)
				{
					latest = devices[i].time;
					gtk_tree_selection_select_iter(selection, &iter);
				}

				++n_listed;
			}

	}

	if(n_listed)
		gtk_widget_show_all(window);
	else
	{
		GtkWidget *dialog;

		dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
			"No devices to %s", (umount ? "unmount" : "mount"));
		g_signal_connect(dialog, "response", G_CALLBACK(&gtk_main_quit), NULL);
		gtk_widget_show_all(dialog);
	}

	gtk_main();

	free_devices(devices);

	return 0;
}
